//******************************************************************
//
//	MMC File Select and Bootloader for "LED Game for AVR"
//
//	MCU:ATmega168
//		efuse:0xF8	u[g̈ 1024word ZbgxN^̓u[g̈
//		hfuse:0xD5	`bv EEPROMی, BOD 2.7V
//		lfuse:0xE2	CR 8MHz
//
//	Compiler:	AVR-GCC		3.4.6
//				AVR-LIBC	1.4.4
//				binutils	2.16.1
//
//	2007.2.4	KAWAKAMI Yukio
//
//	New BSD license
//******************************************************************

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<avr/boot.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

#include	"MMCboot.h"
#include	"sfont.h"

uchar	DiskCache[512];	// fBXNLbV

uchar	FATtype;			// 1:FAT12  2:FAT16
uchar	SectorsPerCluster;	// 1NX^̃ZN^
ushort	ReservedSectors;	// \ZN^
uchar	FATcount;			// FAT̐ ʏ 2
uint	RootDirEntriesCount;	// [gfBNgGg[
ulong	SectorCount;	// ZN^
ushort	FATsectors;		// FATggZN^
uint	FATstart;		// FATJnZN^
uint	DIRstart;		// [gfBNgJnZN^
ulong	DataStart;		// f[^̈JnZN^
//ushort	ClusterCount;	// NX^

ushort	LEDbuff[16];
uchar	LEDrowCount;

volatile uchar SwitchFlag;


//****************************************************************************
// JIvV --section-start=.text=0x3800 Ǝw肷̂
// ʂ̃R[h 0x3800(u[g[_̈)ɔzu
//****************************************************************************

//****************************************************************
// ZtvO~OEC
//
//	̒ŃvȌŝŁA
//	K AVR̃u[g[_̈ɔzu邱
//****************************************************************
uchar spm_main(struct FILE_DATA *fd){
	uint	i;
//	uchar	dot;
	ushort	w;
	ushort	sum;
	ulong	page;

//	dot = 0;

	// EEPROMݏI҂
//	read_eeprom(0);

	page = 0;
	sum = 0;

	while(page < fd->size){
		// MMC/SDPZN^ǂݍ
		if (mmc_read(fd->sec)){
			return 1;
		}
		for (i = 0; i < 512; i += 2){
			if ((i & (SPM_PAGESIZE - 1)) == 0){
				// y[W
				boot_page_erase(page);
				boot_spm_busy_wait();
				sum = 0xFFFF;
			}
			w = *(ushort *)(&DiskCache[i]);
			sum &= w;
			boot_page_fill(page + i, w);
			if ((i & (SPM_PAGESIZE - 1)) == (SPM_PAGESIZE - 2)){
				if (sum != 0xFFFF){	// S FFFF Ȃ珑܂Ȃ
					// y[W
					boot_page_write(page);
					boot_spm_busy_wait();
				}
				page += SPM_PAGESIZE;
			}
		}

//		LED_set(dot & 15, dot >> 4);
//		dot++;
		next_sector(fd);
	}

	// RWWZNVLɂ
	boot_rww_enable();

	// 0ԒnɃWv(RESET)
//	asm volatile("jmp	0\n"::);

	return 0;
}



//*******************************************************
// MMCJ[hPZN^ǂݍ
//*******************************************************
uchar mmc_read(ulong addr){
	int		i;
	uchar	r;
	uchar	*buff;

	buff = DiskCache;
	if (!SPI_read_open(addr)){
		for (i = 0; i < 512; i++){
			*buff++ = SPI_in();
		}
		SPI_read_close();
		r = 0;
	} else {
		r = 1;
	}
	return r;
}


//**********************************************
//	̃ZN^w
//**********************************************
void next_sector(struct FILE_DATA *fd){

	fd->sec += 512;
	fd->p += 512;
	if (++(fd->sc) >= SectorsPerCluster){
		fd->cluster = next_cluster(fd->cluster);
		fd->sec = cluster_to_adrs(fd->cluster);
		fd->sc = 0;
	}
}


//*********************************************************
//	NX^ԍf[^i[AhX擾
//*********************************************************
ulong cluster_to_adrs(uint cluster){

	return (((ulong)cluster - 2) * (ulong)SectorsPerCluster + DataStart) * 512;
}


//*********************************************************
//	̃NX^߂
//*********************************************************
uint next_cluster(uint c){
	ulong	addr;
	uint	data;
	ulong	fatadr;

	if (FATtype == FAT12){
		// FAT12
		addr = (ulong)FATstart * 512 + (ulong)c + (c >> 1);
		if ((addr & 511) == 511){
			// ZN^܂
			data = read_word(addr) & 0x00FF;
			data |= ((read_word(addr+1) & 0x00FF) << 8);
		} else {
			fatadr = addr & 0xFFFFFE00;
			data = read_word(addr);
		}
		if (c & 1){
			c = data >> 4;
		} else {
			c = data & 0x0fff;
		}
		if (c >= 0x0ff8)	c = 0xFFFF;
	} else {
		// FAT16
		addr = (ulong)FATstart * 512 + (ulong)c * 2;
		fatadr = addr & 0xFFFFFE00;
		c = read_word(addr);
		if (c >= 0xFFF8)	c = 0xFFFF;
	}

	return c;
}


//*******************************************************
//	MMC/SDJ[h 2byteǂ
//	ZN^ẼANZX͂łȂ̂
//	ʍHōl邱
//*******************************************************
uint read_word(ulong addr){
	uint	w;
	uint	b;

	b = addr & 511;
	addr &= 0xFFFFFE00;

	mmc_read(addr);

	w = (uint)DiskCache[b++];
	w |= ((uint)DiskCache[b] << 8);

	return w;
}


//****************************************
//	MMC/SD SPIR}ho
//****************************************
uchar SPI_command(uchar com, ulong arg){
	uint	i;
	uchar	r;

	// BUSY`FbN
	for (i = 0; i < 10000; i++){
		r = SPI_in();
		if (r == 0xFF)	break;
	}
	if (r != 0xFF)	return 1;

	SPI_out(com|0x40);
	SPI_out(arg >> 24);
	SPI_out(arg >> 16);
	SPI_out(arg >> 8);
	SPI_out(arg);
	SPI_out(0x95);	// R}hCRC

	// R1҂
	for(i = 0; i < 2000; i++){
		r = SPI_in();
		if (r != 0xFF)	break;
	}

	return r;
}


//****************************************
//	MMC/SDJ[hǂݍ݃I[v
//****************************************
uchar SPI_read_open(ulong adrs){
	uint	i;
	uchar	r;
	uchar	retry;

	// J[hɂĂ̓gCKv̂悤
	for (retry = 0; retry < 240; retry++){
		MMC_CS_ON;
		r = SPI_command(17, adrs);
		if (r == 0){
			for (i = 0; i < 10000; i++){	// ̂҂ꍇ
				r = SPI_in();
				if (r == 0xFE){
					return 0;	// ]Jn}[N
				}
			}
		} else if (r == 0xFE){
			return 0;	// I[v
		}
		MMC_CS_OFF;
		SPI_in();
	}

	return 1;	// I[vłȂ
}


//****************************************
//	MMC/SDJ[hǂݍ݃N[Y
//	1bytêꂪĂCɂȂ
//****************************************
void SPI_read_close(void){
	uint	i;

	SPI_in();	// CRC
	SPI_in();	// CRC
	// BUSY`FbN
	for (i = 0; i < 3000; i++){
		if (SPI_in() == 0xFF)	break;
	}
	MMC_CS_OFF;
	SPI_in();
}


//****************************************
//	SPIf[^o
//****************************************
void SPI_out(uchar data){

	SPDR = data;
	while(!(SPSR & (1<<SPIF)));
}


//****************************************
//	SPIf[^
//****************************************
uchar SPI_in(void){

	SPDR = 0xFF;
	while(!(SPSR & (1<<SPIF)));
	return SPDR;
}


//****************************************************************************
// ȉ̓JIvV --section-start=.bootloader=0x3000 Ǝw肷̂
// 0x3000`ɔzu
//****************************************************************************


//**********************************************
//	^C}[Qrv荞
//	8MHz/32/250=1kHz
//	LED\ƃXCb`Ď
//**********************************************
ISR(TIMER2_COMPA_vect){

	// LED\
	LEDrowCount++;
	LEDrowCount &= 15;

	LED_send_word(0xFFFF);	// LED΂̕
	LED_send_word(LEDbuff[LEDrowCount]);
	LED_send_word(~((ushort)1<<(LEDrowCount)));

	LED_PORT |= (1<<LED_LATCH);
	LED_PORT &= ~(1<<LED_LATCH);

	if (!LEDrowCount){
		// 16msɂPXCb``FbN
		SwitchFlag = (~SW_PORT) &
			((1<<SW_A)|(1<<SW_B)|(1<<SW_UP)|(1<<SW_DN)|(1<<SW_L)|(1<<SW_R));
	}
}


//*******************************************************
//	C[`
//*******************************************************
int main (void) {
	uchar	buff[10];
	struct DIRECTORY dir;
	int		n;
	int		fn;

	cli();	// 荞݋֎~

	// |[gBgpś͓{vAbv
	PORTB = (1<<MMC_MOSI)|(1<<MMC_SCK)|(1<<6)|(1<<7);
	DDRB = (1<<LED_DATA)|(1<<LED_CLOCK)|(1<<LED_LATCH)|(1<<MMC_MOSI)|(1<<MMC_SCK);
	PORTC = 0xFF;	// gpȂ̂ŃvAbv
	DDRC = 0;

	PORTD = (1<<SW_A)|(1<<SW_B)|(1<<SW_UP)|(1<<SW_DN)|(1<<SW_L)|(1<<SW_R);
	DDRD = (1<<SP_OUT)|(1<<MMC_CS);

	// 荞݃xN^u[g̈Ɉړ
	MCUCR = (1<<IVCE);
	MCUCR = (1<<IVSEL);
	
	// ^C}Q CTC  8MHz/32/250=1kHz
	TCCR2A = (1<<WGM21);
	TCCR2B = (1<<CS21)|(1<<CS20);	// CTC vXP[ 1/32
	OCR2A = 250-1;			// ^C}JEgől 250  8MHz/32/250=1kHz
	TIMSK2 = (1<<OCIE2A);	// r`荞݋

	SMCR = (1<<SE);		// X[v ACh[h

	cls();	// ʃNA
	
	sei();	// 荞݋

	// Nɂa{^ĂƕKu[gZN^N
	if (((~SW_PORT) & ((1<<SW_A)|(1<<SW_B))) != (1<<SW_B)){

		start_game();	// AvP[VsB܂ĂȂȂ߂
	}
	
	n = 0;
	if (mmc_init() == 0xFF){	// MMC/SDJ[h
		strcpy_P(buff, PSTR("NO  CARD"));
	} else if (read_VFAT_info()){
		strcpy_P(buff, PSTR("CARDERR."));
	} else {
		n = search_file(&dir, RootDirEntriesCount);
		if (n == 0){
			strcpy_P(buff, PSTR("NO  FILE"));
		} else {
			search_file(&dir, 1);
			strcpy(buff, dir.filename);
		}
	}
	print(buff);

	fn = 1;
	for(;;){
		// XCb`͑҂
		while(SwitchFlag){	SLEEP;	}
		while(!SwitchFlag){	SLEEP;	}

		// `{^͑It@Cu[g
		if ((n > 0)&&(SwitchFlag & (1<<SW_A))){
			if (check_file(&dir)){
				strcpy_P(buff, PSTR("FILEERR."));
			} else {
				// t@Cɖ肪Ȃ̂ŃvO~OJn
				struct FILE_DATA fd;
				cls();
				SLEEP;	// ʂ܂ő҂
				cli();	// 荞݋֎~
				open_file(&dir, &fd);
				if (spm_main(&fd) == 0){
					start_game();	// AvP[Vs
				}
				strcpy_P(buff, PSTR("PROGERR."));
				sei();
			}
			print(buff);
		}

		// a{^́Aɏݍς݂̃vOu[g(LZ)
		if (SwitchFlag & (1<<SW_B)){
			start_game();
		}

		// kAq{^Ńt@CI
		if ((n > 0)&&(SwitchFlag & ((1<<SW_L)|(1<<SW_R)))){
			if (SwitchFlag & (1<<SW_L)){
				if (--fn < 1)	fn = n;
			} else {
				if (++fn > n)	fn = 1;
			}
			search_file(&dir, fn);
			cls();
			print(dir.filename);
		}
	}
}


//**********************************************
//	[U[AvP[VsJn
//**********************************************
void start_game(void){
	ushort code;

	// R[h`FbNBŏ jmp߂
	code = pgm_read_word(0);
	if ((code & 0xFE0E) != 0x940C)	return;

	cli();

	// 荞݃xN^߂
	MCUCR = (1<<IVCE);
	MCUCR = 0;
	
	// ^C}Q~
	TCCR2A = 0;
	TCCR2B = 0;
	TIMSK2 = 0;
	TIFR2 = (1<<OCF2A);	// 荞݃tONA

	// X[v[hftHg
	SMCR = 0;

	// SPI[h
	SPCR = 0;
	SPSR = 0;

	// 0ԒnɃWv
	asm volatile("jmp	0\n"::);
}


//**********************************************
//	t@CI[v
//**********************************************
void open_file(struct DIRECTORY *dir, struct FILE_DATA *fd){

	fd->size = (ushort)dir->size;
	fd->sc = 0;
	fd->cluster = dir->cluster;
	fd->sec = cluster_to_adrs(fd->cluster);
	fd->p = 0;
}


//*******************************************************
//	ȒPȃt@C`FbN
//*******************************************************
uchar check_file(struct DIRECTORY *dir){
	ulong	sec;

	if (dir->size > (12 * 1024))	return 1;	// TCY 12KB傫

	// ŏ̃ZN^ǂ
	sec = cluster_to_adrs(dir->cluster);
	if (mmc_read(sec))	return 2;				// ǂݍ݃G[

	// ŏ jmp߂(Flashe 16KBȏ AVȐꍇ jmp߂ɂȂ)
	if (((DiskCache[0] & 0x0E) != 0x0C)||
		((DiskCache[1] & 0xFE) != 0x94)){
		return 3;	// f[^G[
	}

	return 0;
}


//*********************************************************
//	MMC/SDJ[h̏
//*********************************************************
uchar mmc_init(void){
	uchar	c;

	// SPI}X^, CPOL=1, CPHA=1, Clock 8MHz/64=125kHz
	SPCR = (1<<SPE)|(1<<MSTR)|(1<<CPOL)|(1<<CPHA)|(1<<SPR0)|(1<<SPR1);
	SPSR = (1<<SPI2X);

	MMC_CS_OFF;	// CS=H

	// MMC/SD SPI[hڍsÕNbNM
	for (c=0; c < 20; c++){
		SPI_in();
	}

	// SPI[hڍsR}h
	MMC_CS_ON;	// CS=L
	c = SPI_command(0, 0);
	while ((c != 1)&&(c != 0xFF)){
		c = SPI_command(0, 0);
	}
	while (c == 1){
		c = SPI_command(1, 0);
	}
	MMC_CS_OFF;	// CS=H
	SPI_in();

	// SPI}X^, CPOL=1, CPHA=1, Clock 8MHz/2=4MHz
	SPCR = (1<<SPE)|(1<<MSTR)|(1<<CPOL)|(1<<CPHA);

	return c;
}


//*********************************************************
//	VFATǂݎ
//*********************************************************
uchar read_VFAT_info(void){
	uchar	buff[32];
	ulong	sec;
	ushort	bpb;
	ushort	n;

	// VFATǂݎBZN^ 512bytesŌߑłĂ
	read_32(446, buff);								// MBRǂݍ
	bpb = *((ushort *)&buff[8]);					// BPB擪ZN^

	read_32((ulong)bpb * 512, buff);				// ǂݍ
	n = *((ushort *)&buff[11]);						// ZN^
	if (n != 512){
		return 2;	// 1ZN^ 512bytesłȂ
	}

	FATcount = buff[16];		// FAT̐ ʏ 2

	SectorsPerCluster = buff[13];		// NX^̃ZN^
	ReservedSectors = *((ushort *)&buff[14]);	// \ZN^
	FATstart = bpb + ReservedSectors;			// FATJnZN^

	RootDirEntriesCount = *((ushort *)&buff[17]);	// [gfBNgGg

	FATsectors = *((ushort *)&buff[22]);		// FAT̃ZN^
	if (FATsectors == 0){
		return 1;	// FAT32炵
	}

	DIRstart = (ushort)FATstart + (ushort)FATcount * FATsectors;	// FAT̐~ZN^
	n = RootDirEntriesCount / 16;	// [gfBNgɕKvȃZN^
	if (RootDirEntriesCount % 16){
		n++;	// Ô߁A]肪o؂グBʏ͕svƎv
	}
	DataStart = (ulong)DIRstart + (ulong)n;		// f[^i[̈

	sec = (ulong)*((ushort *)&buff[19]);		// ZN^
	if (sec == 0){
		read_32((ulong)bpb * 512 + 32, buff);	// ǂݍ
		sec = *((ulong *)&buff[0]);
	}
	SectorCount = sec;
	sec -= FATstart;
	sec /= SectorsPerCluster;
	if (sec >= 0x1000){
		FATtype = FAT16;	// FAT16
	} else {
		FATtype = FAT12;	// FAT12
	}

	// f[^̈ZN^
//	sec = SectorCount - ReservedSectors - FATsectors * FATcount - (ulong)RootDirEntriesCount * 32 / 512;

	// NX^
//	ClusterCount = (ushort)(sec / SectorsPerCluster) + 2;

	return 0;
}


//*******************************************************
//	MMC/SDJ[h 32byteǂ buffɊi[
//	DiskCachegƂɒ
//*******************************************************
uchar read_32(ulong addr, uchar *buff){
	uint	b;
	uchar	i;

	b = addr & 511;
	addr &= 0xFFFFFE00;
	
	if (mmc_read(addr))	return 1;

	for (i = 0; i < 32; i++){
		*buff++ = DiskCache[b++];
	}

	return 0;
}


//*********************************************************
//	[gfBNg̃t@CT[`
//	gq BIN ̒ʏt@C
//*********************************************************
int search_file(struct DIRECTORY *dir, int max){
	ulong	adrs;
	int		n, i;

	adrs = (ulong)DIRstart * 512;
	n = 0;
	for (i = 0; i < RootDirEntriesCount; i++){
		read_32(adrs, (uchar *)dir);
		adrs += 32;
		if (dir->filename[0] == 0)	break;
		if ((dir->filename[0] != 0xE5)&&
			(!(dir->att & 0x18))&&
			(dir->ext[0] == 'B')&&
			(dir->ext[1] == 'I')&&
			(dir->ext[2] == 'N')){
			if (++n >= max)	break;
		}
	}
	return n;
}


//****************************************
//	SW܂ő҂
//****************************************
void wait_switch_off(void){

	for(;;){
		SLEEP;
		if (!SwitchFlag)	break;
	}
}


//*******************************************************
//	ʃNA
//*******************************************************
void cls(void){
	uchar	i;

	for (i = 0; i < 16; i++){
		LEDbuff[i] = 0;
	}
}


//*******************************************************
//	܂ʒuɕ\
//*******************************************************
void print(uchar *str){
	uchar	c;
	uchar	x;
	uchar	y;

	x = 0;	// W
	y = 2;
	while((c = *str++)){
		if (x > 12){
			x = 0;
			y += 6;
			if (y > 12)	break;
		}
		putc(x, y, c);
		x += 4;
	}
}


//*******************************************************
//	LEDɂP\
//*******************************************************
void putc(uchar x, uchar y, uchar c){
	ushort	font;
	uchar	i;

	font = (ushort)(c - 0x20) * 3;

	for (i = 0; i < 6; i++){
		c = pgm_read_byte(&smallfont[font]);
		if ((i & 1) == 0){
			c >>= 4;
		} else {
			c &= 0x0F;
			font++;
		}
		LEDbuff[y] &= ~(0x000F << (12 - x));
		LEDbuff[y] |= ((ushort)c << (12 - x));
		y++;
	}
}


//*******************************************************
//	Phbg_
//*******************************************************
void LED_set(uchar x, uchar y){
	LEDbuff[y] |= 1<<(15 - x);
}


//*******************************************************
//	LEDɂP[h(16bit)]
//*******************************************************
void LED_send_word(ushort data){
	uchar	count;

	for (count = 0; count < 16; count++){
		if (data & 1){
			LED_PORT |= (1<<LED_DATA);
		} else {
			LED_PORT &= ~(1<<LED_DATA);
		}
		data >>= 1;
		LED_PORT |= (1<<LED_CLOCK);
		LED_PORT &= ~(1<<LED_CLOCK);
	}
}

